#-------------------------------------------------------------------------------
# elftools example: dwarf_die_tree.py
#
# In the .debug_info section, Dwarf Information Entries (DIEs) form a tree.
# pyelftools provides easy access to this tree, as demonstrated here.
#
# Eli Bendersky (eliben@gmail.com)
# This code is in the public domain
#-------------------------------------------------------------------------------
import logging
logger = logging.getLogger(__name__)

import sys

# If pyelftools is not installed, the example can also run from the root or
# examples/ dir of the source distribution.
sys.path[0:0] = ['.', '..']

import collections
from elftools.dwarf.dwarf_expr import GenericExprVisitor
from elftools.dwarf.dwarfinfo import DebugSectionDescriptor
from elftools.dwarf.descriptions import describe_attr_value
import elftools.common.exceptions
import io
import itertools
import os
import epyqlib.ticoff
import traceback

import attr
import bitstruct
import enum
import itertools
import textwrap

import epyqlib.twisted.cancalibrationprotocol as ccp

bits_per_byte = 16

@enum.unique
class TypeFormats(enum.Enum):
    # http://www.dwarfstd.org/doc/Dwarf3.doc
    address = 0x01
    boolean = 0x02
    complex_float = 0x03
    float = 0x04
    signed = 0x05
    signed_char = 0x06
    unsigned = 0x07
    unsigned_char = 0x08
    imaginary_float = 0x09
    packed_decimal = 0x0a
    numeric_string = 0x0b
    edited = 0x0c
    signed_fixed = 0x0d
    unsigned_fixed = 0x0e
    decimal_float = 0x0f
    lo_user = 0x80
    hi_user = 0xff

    def is_integer(self):
        return self.is_signed_integer() or self.is_unsigned_integer()

    def is_signed_integer(self):
        return self in [TypeFormats.signed, TypeFormats.signed_char]

    def is_unsigned_integer(self):
        return self in [TypeFormats.unsigned, TypeFormats.unsigned_char]

    def is_floating_point(self):
        return self in [TypeFormats.float]


class BytesProxy:
    @property
    def bytes(self):
        return self.type.bytes


def bytearray_to_bits(data):
    return ''.join(['{:08b}'.format(b) for b in data])


def process_expression(expression, dwarf_info):
    generic_expression_visitor = GenericExprVisitor(dwarf_info.structs)
    generic_expression_visitor.process_expr(expression)
    result, = generic_expression_visitor._cur_args
    return result


@attr.s
class Type:
    name = attr.ib()
    bytes = attr.ib()
    format = attr.ib()

    # def format_string(self):
    #     if self.format.is_integer():
    #         if self.format.is_unsigned_integer():
    #             type = 'u'
    #         else:
    #             type = 's'
    #     elif self.format.is_floating_point():
    #         # TODO: CAMPid 097897541967932453154321546542175421549
    #
    #         # note that this is hardcoded for two bytes/address fudge
    #         types = {
    #             2: 'f',
    #             4: 'd'
    #         }
    #         try:
    #             type = types[self.bytes]
    #         except KeyError:
    #             raise Exception(
    #                 'float type only supports lengths in [{}]'.
    #                 format(', '.join([str(t) for t in types.keys()]))
    #             )
    #     else:
    #         raise Exception('Unsupported type format: {}'.format(self.type))
    #
    #     return '{}{}{}'.format('>', type, self.bytes * bits_per_byte)

    # TODO: CAMPid 034173541438600605430541538
    def unpack(self, data):
        if isinstance(data, bytearray):
            bits = bytearray_to_bits(data)
        else:
            bits = data

        # TODO: the % seems fishy
        if self.bytes > 1 and len(bits) % bits_per_byte == 0:
            # TODO: CAMPid 08793287728743824372437983526631513679
            bits = ''.join(itertools.chain(*reversed(list(
                grouper(bits, bits_per_byte)))))

        if self.format.is_integer():
            if self.format.is_unsigned_integer():
                pad = '0'
                type = 'u'
            else:
                pad = bits[0]
                type = 's'

            padding = pad * (self.bytes * bits_per_byte - len(bits))

            padded_bits = padding + bits

            format = '>{}{}'.format(type, str(len(bits)))

            if len(padding) > 0:
                format = '>u{}{}'.format(str(len(padding)), format)

            return bitstruct.unpack(
                format,
                int(padded_bits, 2).to_bytes(
                    self.bytes * bits_per_byte // 8,
                    byteorder='big')
            )[-1]

        elif self.format.is_floating_point():
            # TODO: CAMPid 097897541967932453154321546542175421549

            # note that this is hardcoded for two bytes/address fudge
            types = {
                4: 'f',
                8: 'd'
            }
            try:
                type = types[self.bytes * (bits_per_byte / 8)]
            except KeyError:
                raise Exception(
                    'float type only supports lengths in [{}]'.
                    format(', '.join([str(t) for t in types.keys()]))
                )

            format = '{}{}{}>'.format('>', type, self.bytes * bits_per_byte)

            return bitstruct.unpack(
                format,
                bytes(
                    int(''.join(b), 2)
                    for b in epyqlib.utils.general.grouper(bits, 8)
                ),
            )[-1]

        raise Exception('Unsupported type format: {}'.format(self.type))


@attr.s
class UnspecifiedType:
    name = attr.ib()


# consider https://repl.it/EyFY/26
@attr.s
class SubroutineType:
    return_type = attr.ib(default=None)
    name = attr.ib(
        default=None,
        converter=lambda v: str(v) if v is not None else '<subroutine>'
    )
    parameters = members = attr.ib(default=attr.Factory(list))


@attr.s
class PointerType:
    type = attr.ib()
    modifier = '*'

    # note that this is hardcoded and should be detected
    # DW_AT_address_class happens to be 32 but the definition
    # doesn't seem to suggest that means 32-bits
    bytes = 32 // bits_per_byte

    @property
    def name(self):
        return '{}*'.format(type_name(self))

    def format_string(self):
        return '<u{}'.format(self.bytes * bits_per_byte)


    # TODO: CAMPid 034173541438600605430541538
    def unpack(self, data):
        if isinstance(data, bytearray):
            bits = bytearray_to_bits(data)
        else:
            bits = data

        total_bit_count = self.bytes * bits_per_byte

        # TODO: the % seems fishy
        if self.bytes > 1 and len(bits) % bits_per_byte == 0:
            # TODO: CAMPid 08793287728743824372437983526631513679
            bits = ''.join(itertools.chain(*reversed(list(
                grouper(bits, bits_per_byte)))))

        pad = '0'
        type = 'u'

        self.value, = bitstruct.unpack(
            '>' + type + str(total_bit_count),
            int(bits, 2).to_bytes(
                self.bytes * bits_per_byte // 8,
                byteorder='big')
        )

        return self.value
        # bits.extend(pad * (total_bit_count - len(bits)))


@attr.s
class VolatileType(BytesProxy):
    type = attr.ib()
    name = attr.ib(default=None)
    modifier = 'volatile'


@attr.s
class ArrayType:
    type = attr.ib()
    bytes = attr.ib()
    dimensions = attr.ib()
    name = attr.ib(default=None)

    def array_markup(self):
        return '[{}]'.format(self.length())

    def offset_of(self, *indexes):
        offset = 0
        overall_multiplier = self.type.bytes
        x = itertools.zip_longest(indexes, self.dimensions, fillvalue=0)
        x = reversed(list(x))
        for index, multiplier in x:
            offset += overall_multiplier * index
            overall_multiplier *= multiplier

        return offset

    def unpack(self, data):
        # TODO: CAMPid 078587996542145215432667431535465465421
        if isinstance(data, bytes):
            data = bytearray(data)
        elif isinstance(data, str):
            data = bytearray(int(data, 2)
                             .to_bytes(self.bytes * bits_per_byte // 8,
                                       byteorder='big'))
        bits = bytearray_to_bits(data)

        base = base_type(self.type)
        type_bit_size = base.bytes * bits_per_byte

        expected_items = self.length()

        values = []
        for group in grouper(bits, type_bit_size):
            expected_items -= 1
            values.append(base.unpack(''.join(group)))

        if expected_items != 0:
            raise Exception('wrong amount of data for array')

        return values


@attr.s
class ConstType(BytesProxy):
    type = attr.ib()
    name = attr.ib(default=None)
    modifier = 'const'


@attr.s
class RestrictType:
    type = attr.ib()
    name = attr.ib(default=None)


def name(type):
    if hasattr(type, 'name'):
        return type.name
    else:
        return name(type.type)


def full_type(type):
    if isinstance(type, (TypeDef, Type, Union)):
        return type.name
    elif hasattr(type, 'modifier'):
        return '{} {}'.format(type.modifier, full_type(type.type))
    else:
        return full_type(type.type)


def base_type(type):
    if isinstance(type, (PointerType, ArrayType, EnumerationType)):
        return type
    elif hasattr(type, 'type'):
        return base_type(type.type)
    else:
        return type


def type_name(type):
    name = None
    while type is not None:
        if not hasattr(type, 'type'):
            name = type.name
            break

        type = type.type

        if hasattr(type, 'name'):
            if type.name is not None and len(type.name) > 0:
                name = type.name
                break

    return name


def location(member):
    l = member.location
    if member.bit_offset is not None:
        l = '{}:{}'.format(l, member.bit_offset)
    return l


@attr.s
class Struct:
    bytes = attr.ib()
    name = attr.ib(default=None)
    members = attr.ib(default=attr.Factory(collections.OrderedDict))

    def render(self, terminate=True):
        members = []
        for member in self.members.values():
            members.append(
                '    {type} {name}{bits}; // {base_type} {size}@{location}'
                    .format(
                        type=full_type(member),
                        name=member.name,
                        bits=':{}'.format(member.bit_size)
                             if member.bit_size is not None else '',
                        base_type=base_type(member).name,
                        size=base_type(member).bytes,
                        location=location(member)
                    )
            )
        return textwrap.dedent('''\
        struct {name}{{
        {members}
        }}{terminator} /* {size} */''').format(
            name=self.name + ' ' if self.name is not None else '',
            members='\n'.join(members),
            terminator=';' if terminate else '',
            size=self.bytes)

    # def format_string(self):
    #     format = ''
    #     for member in self.padded_members():
    #         if member.name == '<padding>':
    #             f = '<p{}'.format(member.bit_size)
    #         elif member.bit_offset is None:
    #             f = member.format_string()
    #         else:
    #             f = '>{}{}'.format(member.format_string()[1], member.bit_size)
    #
    #         format += f
    #
    #     return format

    def padded_members(self):
        locations = collections.defaultdict(list)
        for member in self.members.values():
            if member.bit_offset is None:
                range = (
                    0,
                    (base_type(member).bytes * bits_per_byte) - 1,
                    member
                )
            else:
                range = (
                    member.bit_offset,
                    member.bit_offset+member.bit_size-1,
                    member
                )

            locations[member.location].append(range)

        for location, ranges in locations.items():
            ranges.sort()
            padding = []
            bit = 0
            for range in ranges:
                if range[0] > bit:
                    pad = StructMember(
                        name='<padding>',
                        type=range[2].type,
                        location=location,
                        bit_offset=bit,
                        bit_size=range[0] - bit,
                        padding=True
                    )

                    padding.append((bit, range[0] - 1, pad))
                bit = range[1] + 1
            ranges.extend(padding)
            ranges.sort()
            ranges[:] = [r[2] for r in ranges]

        ranges = []
        for location in sorted(locations.keys()):
            ranges.extend(locations[location])

        return ranges

    def unpack(self, data):
        # TODO: CAMPid 078587996542145215432667431535465465421
        if isinstance(data, bytes):
            data = bytearray(data)
        elif isinstance(data, str):
            data = bytearray(int(data, 2)
                             .to_bytes(self.bytes * bits_per_byte // 8,
                                       byteorder='big'))

        location = None
        for member in self.padded_members():
            b_type = base_type(member)
            if location != member.location:
                location = member.location
                if b_type.bytes > 1 and isinstance(b_type, Type) and member.bit_size is not None:
                    # TODO: CAMPid 08793287728743824372437983526631513679
                    byte = member.location * bits_per_byte // 8
                    s = slice(byte, byte + b_type.bytes * bits_per_byte // 8)
                    data[s] = bytearray(itertools.chain(*reversed(list(
                        grouper(data[s], bits_per_byte // 8)))))

        bits = bytearray_to_bits(data)
        values = collections.OrderedDict()
        for member in self.padded_members():
            if member.name != '<padding>':
                bit_size = (member.bit_size
                            if member.bit_size is not None
                            else member.bytes * bits_per_byte)

                bit_offset = (member.bit_offset
                              if member.bit_offset is not None
                              else 0)

                msb = (member.location * bits_per_byte) + bit_offset
                lsb = msb + bit_size

                values[member.name] = member.unpack(bits[msb:lsb])

        # return collections.OrderedDict(reversed(list(values.items())))

        # raw_values = bitstruct.unpack(self.format_string(), data)
        #
        # d = dict(
        #     zip(
        #         (m.name for m in self.padded_members() if m.name != '<padding>'),
        #         raw_values
        #     )
        # )
        #
        return collections.OrderedDict(
            (m.name, values[m.name]) for m in self.members.values())

    # TODO: refactor to *args
    def offset_of(self, names):
        offset = 0

        struct = self

        for name in names:
            member = struct.members[name]
            offset += member.location
            struct = base_type(member)

        return offset

    def member(self, names):
        struct = self

        for name in names:
            member = base_type(struct.members[name])
            struct = member

        return member

    def member_and_offset(self, names):
        return self.member(names), self.offset_of(names)

# https://docs.python.org/3/library/itertools.html
def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)


@attr.s
class StructMember:
    name = attr.ib()
    type = attr.ib()
    location = attr.ib()
    bit_offset = attr.ib(default=None)
    bit_size = attr.ib(default=None)
    padding = attr.ib(default=False)

    def format_string(self):
        format = base_type(self).format_string()
        # format = '>' + format[1:]
        return format

    @property
    def bytes(self):
        return base_type(self).bytes

    def unpack(self, data):
        if isinstance(data, bytearray):
            bits = bytearray_to_bits(data)
            if self.bit_size is not None:
                bits = bits[self.bit_offset:]
                bits = bits[:self.bit_size]
        else:
            bits = data

        return base_type(self).unpack(bits)

@attr.s
class Union:
    bytes = attr.ib()
    name = attr.ib(default=None)
    members = attr.ib(default=attr.Factory(collections.OrderedDict))

    def unpack(self, data):
        # # TODO: CAMPid 078587996542145215432667431535465465421
        # if isinstance(data, bytes):
        #     data = bytearray(data)
        # elif isinstance(data, str):
        #     data = bytearray(int(data, 2)
        #                      .to_bytes(self.bytes * bits_per_byte // 8,
        #                                byteorder='big'))

        values = collections.OrderedDict()
        for member in self.members.values():
            b_type = base_type(member)
            values[member.name] = member.unpack(data)

        return collections.OrderedDict(
            (m.name, values[m.name])
            for m in self.members.values()
        )


@attr.s
class UnionMember:
    name = attr.ib()
    type = attr.ib()
    # location = attr.ib()
    # bit_offset = attr.ib(default=None)
    # bit_size = attr.ib(default=None)
    # padding = attr.ib(default=False)

    def format_string(self):
        format = base_type(self).format_string()
        # format = '>' + format[1:]
        return format

    @property
    def bytes(self):
        return base_type(self).bytes

    def unpack(self, data):
        return base_type(self).unpack(data)
        # if isinstance(data, bytearray):
        #     bits = bytearray_to_bits(data)
        # else:
        #     bits = data
        #
        # if self.bit_size is not None:
        #     bits = bits[self.bit_offset:]
        #     bits = bits[:self.bit_size]
        # return base_type(self).unpack(bits)


@attr.s
class PointerToMember:
    name = attr.ib()


@attr.s
class EnumerationType:
    bytes = attr.ib()
    name = attr.ib(default=None)
    type = attr.ib(default=None)
    values = attr.ib(default=attr.Factory(list))

    # TODO: CAMPid 034173541438600605430541538
    def unpack(self, data):
        if isinstance(data, bytearray):
            bits = bytearray_to_bits(data)
        else:
            bits = data

        total_bit_count = self.bytes * bits_per_byte

        # TODO: the % seems fishy
        if self.bytes > 1 and len(bits) % bits_per_byte == 0:
            # TODO: CAMPid 08793287728743824372437983526631513679
            bits = ''.join(itertools.chain(*reversed(list(
                grouper(bits, bits_per_byte)))))

        pad = '0'
        type = 'u'

        self.value, = bitstruct.unpack(
            '>' + type + str(total_bit_count),
            int(bits, 2).to_bytes(
                self.bytes * bits_per_byte // 8,
                byteorder='big')
        )

        return self.value
        # bits.extend(pad * (total_bit_count - len(bits)))


@attr.s
class EnumerationValue:
    name = attr.ib()
    value = attr.ib()


@attr.s
class TypeDef:
    name = attr.ib()
    type = attr.ib()

    # @property
    # def bytes(self):
    #     return self.type.bytes
    #
    # @property
    # def format(self):
    #     return self.type.format

    # @property
    # def base_type(self):
    #     if isinstance(self.type, Type):
    #         return self.type
    #     else:
    #         return self.type.base_type

    def render(self):
        return 'typedef {type} {name};'.format(
            type=self.type.render(terminate=False),
            name=self.name)

    def unpack(self, bits):
        return base_type(self).unpack(bits)

    @property
    def bytes(self):
        return self.type.bytes


@attr.s
class LoUser(BytesProxy):
    type = attr.ib()


@attr.s
class HiUser:
    type = attr.ib()


@attr.s
class Variable:
    name = attr.ib()
    type = attr.ib()
    address = attr.ib()
    file = attr.ib(default=None)

    def render(self):
        base = base_type(self)
        if isinstance(base, ArrayType):
            array = base.array_markup()
        else:
            array = ''
        return '{type} {name}{array}; // {base} @ 0x{address:08X} from {file}'.format(
            type=full_type(self),
            name=name(self),
            array=array,
            base=base_type(self).name,
            address=self.address,
            file=self.file
        )

    def unpack(self, data):
        return base_type(self).unpack(data)


def get_value(variable):
    # TODO: totally stubbed...
    return 42


def is_modifier(type):
    return isinstance(type, (ConstType,
                             #PackedType,
                             PointerType,
                             #ReferenceType,
                             RestrictType,
                             #SharedType,
                             VolatileType))


def dereference(variable):
    name = '&' + variable.name

    type = variable
    while not isinstance(type, PointerType):
        type = type.type

    # one more to get past the pointer
    type = type.type

    address = get_value(variable)

    return Variable(name=name, type=type, address=address)


def fake_section(filename, section_name):

    with open(os.path.splitext(filename)[0] + section_name, 'rb') as f:
        debug_bytes = f.read()

    return DebugSectionDescriptor(
        stream=io.BytesIO(debug_bytes),
        name=section_name,
        global_offset=0,
        size=len(debug_bytes))


def get_die_path(die):
    path = []

    while die._parent is not None:
        die = die._parent

        attribute = die.attributes.get('DW_AT_name')
        if attribute is None:
            name = '<unknown>'
        else:
            name = attribute.value.decode('utf-8')
        name = name.lstrip('./')
        path.append(name)

    return '::'.join(path[::-1])


def process_file(filename):
    logging.debug('Processing file: {}'.format(filename))
    logging.debug('Working directory: {}'.format(os.getcwd()))

    coff = epyqlib.ticoff.Coff()
    coff.from_file(filename)

    section_bytes = {s.name: (io.BytesIO(s.data), len(s.data))
                     for s in coff.sections
                     if s.name.startswith('.debug_')}
    debug_sections = {
        name: DebugSectionDescriptor(
            stream=stream,
            name=name,
            global_offset=0,
            size=length,
            address=0,
        )
        for name, (stream, length) in section_bytes.items()
    }

    from elftools.dwarf.dwarfinfo import DWARFInfo, DwarfConfig
    dwarfinfo = DWARFInfo(
        config=DwarfConfig(
            little_endian=True,
            default_address_size=4,
            machine_arch='<unknown>'),
        debug_info_sec=debug_sections.get('.debug_info', None),
        # debug_info_sec=DebugSectionDescriptor(
        #     stream=io.BytesIO(dwarf_debug_info_bytes),
        #     name='.debug_info',
        #     global_offset=0,
        #     size=len(dwarf_debug_info_bytes)),
        debug_aranges_sec=debug_sections.get('.debug_aranges', None),
        debug_abbrev_sec=debug_sections.get('.debug_abbrev', None),
        debug_frame_sec=debug_sections.get('.debug_frame', None),
        # TODO(eliben): reading of eh_frame is not hooked up yet
        eh_frame_sec=None,
        debug_str_sec=debug_sections.get('.debug_str', None),
        debug_loc_sec=debug_sections.get('.debug_loc', None),
        debug_ranges_sec=debug_sections.get('.debug_ranges', None),
        debug_line_sec=debug_sections.get('.debug_line', None),
        debug_pubtypes_sec=debug_sections.get('.pubtypes_sec', None),
        debug_pubnames_sec=debug_sections.get('.pubnames_sec', None),
    )

    objects = collections.OrderedDict((tag, []) for tag in [
        'DW_TAG_subprogram',
        'DW_TAG_variable',
        'DW_TAG_typedef',
        'DW_TAG_base_type',
        'DW_AT_encoding',
        'DW_TAG_structure_type',
        'DW_TAG_union_type',
        'DW_TAG_ptr_to_member_type',
        'DW_TAG_enumeration_type',
        'DW_TAG_pointer_type',
        'DW_TAG_array_type',
        'DW_TAG_volatile_type',
        'DW_TAG_const_type',
        'DW_TAG_restrict_type',
        'DW_TAG_lo_user',
        'DW_TAG_hi_user',
        'DW_TAG_unspecified_type',
        'DW_TAG_subroutine_type'
    ])

    for CU in dwarfinfo.iter_CUs():
    # it = dwarfinfo.iter_CUs()
    # while True:
    #     try:
    #         CU = next(it)
    #     except StopIteration:
    #         break
    #     except elftools.common.exceptions.DWARFError:
    #         traceback.print_exc()
    #         logging.debug('Skipping current CU')
    #         next

        # DWARFInfo allows to iterate over the compile units contained in
        # the .debug_info section. CU is a CompileUnit object, with some
        # computed attributes (such as its offset in the section) and
        # a header which conforms to the DWARF standard. The access to
        # header elements is, as usual, via item-lookup.
        logging.debug('  Found a compile unit at offset %s, length %s' % (
            CU.cu_offset, CU['unit_length']))

        # Start with the top DIE, the root for this CU's DIE tree
        top_DIE = CU.get_top_DIE()
        logging.debug('    Top DIE with tag=%s' % top_DIE.tag)

        path = top_DIE.get_full_path()
        # We're interested in the filename...
        logging.debug('    name=%s' % path)

        if path.endswith('__TI_internal'):
            logging.debug('__TI_internal found, terminating DWARF parsing')
            break
        else:
            # Display DIEs recursively starting with top_DIE
            die_info_rec(top_DIE, objects=objects)
            # pass

    def die_info_rec_structure_type(die, indent_level):
        for child in die.iter_children():
            # logging.debug(indent_level + str(child.attributes['DW_AT_name'].value.decode('utf-8')))
            location = str(child.attributes['DW_AT_data_member_location'].value)
            name = str(child.attributes['DW_AT_name'].value.decode('utf-8'))
            logging.debug(indent_level + name + ': ' + location)
            # logging.debug(indent_level + str(child.attributes['DW_AT_name'].value.decode('utf-8')) + ': ' + str(child.attributes['DW_AT_data_member_location'].value.decode('utf-u')))

    # this is yucky but the embedded system is weird with two bytes
    # per address and even sizeof() responds in units of addressable units
    # rather than actual bytes
    byte_size_fudge = 1

    offsets = {}

    types = []
    for die in objects['DW_TAG_base_type']:
        type = Type(
            name=die.attributes['DW_AT_name'].value.decode('utf-8'),
            bytes=die.attributes['DW_AT_byte_size'].value * byte_size_fudge,
            format=TypeFormats(die.attributes['DW_AT_encoding'].value)
        )
        types.append(type)
        offsets[die.offset] = type
        logging.debug('{: 10d} {}'.format(die.offset, type))

    variables = []
    for die in objects['DW_TAG_variable']:
        location = die.attributes.get('DW_AT_location', [])
        if location:
            location = location.value

        # TODO: check this better
        if len(location) != 5:
            continue
        address = process_expression(
            expression=location,
            dwarf_info=dwarfinfo,
        )

        variable = Variable(
            name=die.attributes['DW_AT_name'].value.decode('utf-8'),
            type=die.attributes['DW_AT_type'].value,
            address=address,
            file=get_die_path(die)
        )
        variables.append(variable)
        offsets[die.offset] = variable
        logging.debug('{: 10d} {}'.format(die.offset, variable))

    lo_users = []
    for die in objects['DW_TAG_lo_user']:
        name = die.attributes.get('DW_AT_name', None)
        if name is not None:
            name = name.value.decode('utf-8')
        lo_user = LoUser(
            type=die.attributes['DW_AT_type'].value
        )
        lo_users.append(lo_user)
        offsets[die.offset] = lo_user
        logging.debug('{: 10d} {}'.format(die.offset, lo_user))

    hi_users = []
    for die in objects['DW_TAG_hi_user']:
        name = die.attributes.get('DW_AT_name', None)
        if name is not None:
            name = name.value.decode('utf-8')
        hi_user = HiUser(
            type=die.attributes['DW_AT_type'].value
        )
        hi_users.append(hi_user)
        offsets[die.offset] = hi_user
        logging.debug('{: 10d} {}'.format(die.offset, hi_user))

    subroutine_types = []
    for die in objects['DW_TAG_subroutine_type']:
        name = die.attributes.get('DW_AT_name', None)
        if name is not None:
            name = name.value.decode('utf-8')
        type = die.attributes.get('DW_AT_type', None)
        if type is not None:
            type = type.value
        subroutine_type = SubroutineType(
            name=name,
            return_type=type
        )
        for parameter in die.iter_children():
            subroutine_type.parameters.append(
                parameter.attributes['DW_AT_type'].value)
        subroutine_types.append(subroutine_type)
        offsets[die.offset] = subroutine_type
        logging.debug('{: 10d} {}'.format(die.offset, subroutine_type))

    unspecified_types = []
    for die in objects['DW_TAG_unspecified_type']:
        name = die.attributes.get('DW_AT_name', None)
        if name is not None:
            name = name.value.decode('utf-8')
        unspecified_type = UnspecifiedType(
            name=name
        )
        unspecified_types.append(unspecified_type)
        offsets[die.offset] = unspecified_type
        logging.debug('{: 10d} {}'.format(die.offset, unspecified_type))

    pointer_types = []
    for die in objects['DW_TAG_pointer_type']:
        type = die.attributes['DW_AT_type'].value
        name = die.attributes.get('DW_AT_name', None)
        if name is not None:
            name = name.value.decode('utf-8')
            pointer_type = PointerType(name=name, type=type)
        else:
            pointer_type = PointerType(type=type)
        pointer_types.append(pointer_type)
        offsets[die.offset] = pointer_type
        logging.debug('{: 10d} {}'.format(die.offset, pointer_type))

    volatile_types = []
    for die in objects['DW_TAG_volatile_type']:
        name = die.attributes.get('DW_AT_name', None)
        if name is not None:
            name = name.value.decode('utf-8')
        volatile_type = VolatileType(
            name=name,
            type=die.attributes['DW_AT_type'].value
        )
        volatile_types.append(volatile_type)
        offsets[die.offset] = volatile_type
        logging.debug('{: 10d} {}'.format(die.offset, volatile_type))

    array_types = []
    for die in objects['DW_TAG_array_type']:
        name = die.attributes.get('DW_AT_name', None)
        if name is not None:
            name = name.value.decode('utf-8')
        byte_size = die.attributes.get('DW_AT_byte_size', None)
        if byte_size is not None:
            byte_size = byte_size.value

        dimensions = []
        for child in die.iter_children():
            if child.tag != 'DW_TAG_subrange_type':
                continue

            dimensions.append(child.attributes['DW_AT_upper_bound'].value + 1)

        array_type = ArrayType(
            name=name,
            bytes=byte_size,
            dimensions=dimensions,
            type=die.attributes['DW_AT_type'].value
        )
        array_types.append(array_type)
        offsets[die.offset] = array_type
        logging.debug('{: 10d} {}'.format(die.offset, array_type))
        tags = ('DW_AT_stride_size',)
        for tag_name in tags:
            tag = die.attributes.get(tag_name)
            if tag is not None:
                logging.debug(' found a {}: {}'.format(tag_name, tag))

    const_types = []
    for die in objects['DW_TAG_const_type']:
        name = die.attributes.get('DW_AT_name', None)
        if name is not None:
            name = name.value.decode('utf-8')
        const_type = ConstType(
            name=name,
            type=die.attributes['DW_AT_type'].value
        )
        const_types.append(const_type)
        offsets[die.offset] = const_type
        logging.debug('{: 10d} {}'.format(die.offset, const_type))

    restrict_types = []
    for die in objects['DW_TAG_restrict_type']:
        name = die.attributes.get('DW_AT_name', None)
        if name is not None:
            name = name.value.decode('utf-8')
        restrict_type = RestrictType(
            name=name,
            type=die.attributes['DW_AT_type'].value
        )
        restrict_types.append(restrict_type)
        offsets[die.offset] = restrict_type
        logging.debug('{: 10d} {}'.format(die.offset, restrict_type))

    structure_types = []
    for die in objects['DW_TAG_structure_type']:
        name = die.attributes.get('DW_AT_name', None)
        if name is not None:
            name = name.value.decode('utf-8')
        byte_size_attribute = die.attributes.get('DW_AT_byte_size')
        if byte_size_attribute is None:
            print('Skipping DW_TAG_structure_type due to lack of '
                  'DW_AT_byte_size', name)
            continue
        struct = Struct(
            name=name,
            bytes=byte_size_attribute.value
        )
        structure_types.append(struct)
        offsets[die.offset] = struct
        for member_die in die.iter_children():
            a = member_die.attributes
            bit_offset = a.get('DW_AT_bit_offset', None)
            if bit_offset is not None:
                bit_offset = bit_offset.value
            bit_size = a.get('DW_AT_bit_size', None)
            if bit_size is not None:
                bit_size = bit_size.value
            # TODO: location[1] is just based on observation
            name = a['DW_AT_name'].value.decode('utf-8')

            parsed_location = process_expression(
                expression=a['DW_AT_data_member_location'].value,
                dwarf_info=dwarfinfo
            )

            struct.members[name] = StructMember(
                name=name,
                type=a['DW_AT_type'].value,
                location=parsed_location,
                bit_offset=bit_offset,
                bit_size=bit_size
            )
        logging.debug(list(die.iter_children()))
        logging.debug('{: 10d} {}'.format(die.offset, struct))

    union_types = []
    for die in objects['DW_TAG_union_type']:
        name = die.attributes.get('DW_AT_name', None)
        if name is not None:
            name = name.value.decode('utf-8')
        byte_size_attribute = die.attributes.get('DW_AT_byte_size')
        if byte_size_attribute is None:
            print('Skipping DW_TAG_union_type due to lack of '
                  'DW_AT_byte_size', name)
            continue

        members = collections.OrderedDict((
            (
                member.attributes['DW_AT_name'].value.decode('utf-8'),
                UnionMember(
                    name=member.attributes['DW_AT_name'].value.decode('utf-8'),
                    type=member.attributes.get('DW_AT_type').value,
                ),
            )
            for member in die.iter_children()
        ))

        union = Union(
            name=name,
            bytes=byte_size_attribute.value,
            members=members,
        )
        union_types.append(union)
        offsets[die.offset] = union
        logging.debug('{: 10d} {}'.format(die.offset, union))

    pointer_to_member_types = []
    for die in objects['DW_TAG_ptr_to_member_type']:
        name = die.attributes.get('DW_AT_name', None)
        if name is not None:
            name = name.value.decode('utf-8')
        pointer_to_member = PointerToMember(name=name)
        pointer_to_member_types.append(pointer_to_member)
        offsets[die.offset] = pointer_to_member
        logging.debug('{: 10d} {}'.format(die.offset, pointer_to_member))

    enumeration_types = []
    for die in objects['DW_TAG_enumeration_type']:
        name = die.attributes.get('DW_AT_name', None)
        if name is not None:
            name = name.value.decode('utf-8')
        type = die.attributes.get('DW_AT_type', None)
        if type is not None:
            type = type.value
        enumeration = EnumerationType(
            name=name,
            bytes=die.attributes['DW_AT_byte_size'].value * byte_size_fudge,
            type=type
        )
        for value in die.iter_children():
            enumeration.values.append(
                EnumerationValue(
                    name=value.attributes['DW_AT_name'].value.decode('utf-8'),
                    value=value.attributes['DW_AT_const_value'].value
                )
            )
        enumeration_types.append(enumeration)
        offsets[die.offset] = enumeration
        logging.debug('{: 10d} {}'.format(die.offset, enumeration))
    
    typedefs = []
    for die in objects['DW_TAG_typedef']:
        die_type = die.attributes.get('DW_AT_type')
        if die_type is not None:
            die_type = die_type.value

        typedef = TypeDef(
            name=die.attributes['DW_AT_name'].value.decode('utf-8'),
            type=(die.offset, die_type)
        )
        typedefs.append(typedef)
        offsets[die.offset] = typedef

    offset_values = sorted(offsets.keys())
    logging.debug(len(offset_values))
    logging.debug(offset_values)
    fails = 0
    for typedef in typedefs:
        offset = typedef.type[0]
        try:
            typedef.type = offsets[typedef.type[1]]
        except KeyError:
            logging.debug('Failed to find type for {}'.format(typedef))
            fails += 1
        else:
            logging.debug('{: 10d} {}'.format(offset, typedef))
    logging.debug(fails)

    for structure in structure_types:
        for member in structure.members.values():
            member.type = offsets[member.type]

    for union in union_types:
        for member in union.members.values():
            member.type = offsets[member.type]

    passes = 0
    while True:
        logging.debug('Starting pass {}'.format(passes))
        pass_again = False
        for item in subroutine_types:
            if isinstance(item.return_type, int):
                item.return_type = offsets[item.return_type]
            for i, parameter in enumerate(item.parameters):
                if isinstance(parameter, int):
                    item.parameters[i] = offsets[parameter]

        for item in offsets.values():
            if hasattr(item, 'type') and isinstance(item.type, int):
                try:
                    item.type = offsets[item.type]
                except KeyError:
                    if passes >= 10:
                        logging.debug(item)
                        raise
                    pass_again = True

        passes += 1

        if not pass_again:
            break

    # for pointer_type in pointer_types:
    #     logging.debug(pointer_type)
    #     pointer_type.type = offsets[pointer_type.type]
    #     logging.debug(pointer_type)
    #
    # for array_type in array_types:
    #     logging.debug(array_type)
    #     array_type.type = offsets[array_type.type]
    #     logging.debug(array_type)
    #
    # for volatile_type in volatile_types:
    #     logging.debug(volatile_type)
    #     volatile_type.type = offsets[volatile_type.type]
    #     logging.debug(volatile_type)

    names = collections.defaultdict(list)
    for item in offsets.values():
        if hasattr(item, 'name'):
            valid = False
            if item.name is None:
                valid = True
            elif is_modifier(item):
                pass
            elif item.name.startswith('$'):
                pass
            elif isinstance(item, SubroutineType):
                pass
            else:
                valid = True

            if valid:
                names[item.name].append(item)

    result = names, variables, bits_per_byte

    logging.debug('Finished processing file: {}'.format(filename))

    return result

def testit(names, variables):
    def nonesorter(a):
        if a[0] is None:
            return '', a[1]
        return a

    with open('output.txt', 'w') as f:
        for name, item in sorted(names.items(), key=nonesorter):
            i = item
            logging.debug('{}: {}'.format(name, i), file=f)

    import operator
    logging.debug('\n'.join(str(v) for v in sorted(variables, key=operator.attrgetter('name'))))
    logging.debug(names['Vholdoff'].render())
    logging.debug(names['ozFpga'].render())
    logging.debug(names['FpgaRegs'].render())
    logging.debug(names['writingEE'].render())
    logging.debug(names['rxStart'].render())
    logging.debug(dereference(names['rxStart']).render())
    logging.debug(names['POR_Flags'].render())
    logging.debug(names['IEEE_1547_FreqLimit'].render())
    # logging.debug(base_type(names['IEEE_1547_FreqLimit']).format_string())
    a = b'\x01\x02\x03\x04'
    b = b'\x05\x06\x07\x08'
    c = a + b
    # logging.debug(base_type(names['IEEE_1547_FreqLimit']).unpack(c))
    logging.debug((int.from_bytes(a, byteorder='little'),
           int.from_bytes(b, byteorder='little')))


    data = b'\x00\xd1\x01\x44\x00\x06\x00\x00\x00\x07\x00\x08\x28\x09\x00\xb0'
    # data = bytes(ccp.endianness_swap_2byte(data_raw))#b'\xd1\x00\x44\x01\x06\x00\x00\x00'
    # logging.debug(data)
    # logging.debug(''.join(['{:08b}'.format(b) for b in data]))
    logging.debug(names['TestStruct'].render())
    import pprint
    pp = pprint.PrettyPrinter(indent=4)
    logging.debug(pp.pformat(base_type(names['TestStruct']).padded_members()))
    # logging.debug(base_type(names['TestStruct']).format_string())
    logging.debug(bytearray_to_bits(data))
    logging.debug(pp.pformat(base_type(names['TestStruct']).unpack(data)))

    data = bytearray(data)
    for member in base_type(names['TestStruct']).padded_members():
        type = base_type(member)
        if member.bit_offset is None:
            byte = member.location * bits_per_byte // 8
            s = slice(byte, byte + type.bytes * bits_per_byte // 8)
            # logging.debug('slice: {}'.format(s))
            # logging.debug(data)
            data[s] = ccp.endianness_swap_2byte(data[s])
            # logging.debug(data)

    #                              |  |  |     |  |         |
    logging.debug(pp.pformat(bitstruct.unpack('<p4>u6>u3>u3<p4>u6>u6<u32<p6>u10<u16<p2>u10<u10<u10',
                               data)))


    logging.debug('-----------------------------')
    logging.debug('-----------------------------')
    logging.debug('-----------------------------')

    data = collections.OrderedDict()
    data['TestStruct1'] = b'\x03\xFF\x00\x00'
    data['TestStruct2'] = b'\xFF\xFF\x00\x0F'
    data['TestStruct3'] = b'\xFF\xFF\x3F\xFF'
    data['TestStruct4'] = b'\x03\xFF'
    data['TestStruct5'] = b'\x03\xFF\x00\x00\x00\x00\x00\x00'
    data['TestStruct6'] = b'\xFF\xFF\x00\x0F\x00\x00\x00\x00'
    data['TestStruct7'] = b'\xFF\xFF\x3F\xFF\x00\x00\x00\x00'
    data['TestStruct8'] = b'\xFF\xFF\xFF\xFF\x00\xFF\x00\x00'
    data['TestStruct9'] = b'\xFF\xFF\xFF\xFF\xFF\xFF\x00\x03'
    data['TestStruct10'] = b'\xFF\xFF\xFF\xFF\xFF\xFF\x0F\xFF'
    data['TestStruct11'] = b'\xFF\xFF\xFF\xFF\xFF\xFF\x0F\xFF\x44\x05\x00\x00'
    data['TestStruct12'] = b'\xFF\xFF\xFF\xFF\xFF\xFF\x0F\xFF\x44\x05\x00\x00\x00\x2A\x00\x00\x00\x18\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x04\x00\x00\x00\x05\x00\x00'
    # huh? 6 entries listed in the debugger for a 5 entry array?
    data['testArray1'] = b'\x00\x0B\x00\x00\x00\x16\x00\x00\x00\x21\x00\x00\x00\x2C\x00\x00\x00\x37\x00\x00'#\x00\x05\x35\x4A'

    for name in data.keys():#['TestStruct{}'.format(i) for i in [1,2,3,4,5,6,7,8,9,10,11]]:
        logging.debug(names[name].render())
        base = base_type(names[name])
        if hasattr(base, 'padded_members'):
            logging.debug(pp.pformat(base.padded_members()))
        # logging.debug(base_type(names[name]).format_string())
        logging.debug(bytearray_to_bits(data[name]))
        logging.debug(pp.pformat(base.unpack(data[name])))

        # data = bytearray(data)
        # for member in base_type(names[name]).padded_members():
        #     type = base_type(member)
        #     if member.bit_offset is None:
        #         byte = member.location * bits_per_byte // 8
        #         s = slice(byte, byte + type.bytes * bits_per_byte // 8)
        #         # logging.debug('slice: {}'.format(s))
        #         # logging.debug(data)
        #         data[s] = ccp.endianness_swap_2byte(data[s])
        #         # logging.debug(data)
        #
        # logging.debug(pp.pformat(bitstruct.unpack('<p2<u10<u10<u10',
        #                            data)))
        logging.debug('\n\n-----------------------------\n\n')




    logging.debug('-----------------------------')
    logging.debug('-----------------------------')
    logging.debug('-----------------------------')
    return


    logging.debug(names['Relay'].render())
    # logging.debug(base_type(names['Relay']).bytes)
    # logging.debug(base_type(names['Relay']).format_string())
    data = b'\x00\x48\x00\x00\xd5\x13\x00\x00'
    logging.debug(pp.pformat(data))
    # data = bytes(ccp.endianness_swap_2byte(data))
    # <u2<u1<u2<u1<u1<p9<s16<u32
    # <u1<u1<u2<u1<u2<p9<s16<u32
    # logging.debug(''.join(['{:08b}'.format(b) for b in data]))
    # logging.debug(bitstruct.unpack('<p1<u1<u1<u2<u1<u2<p8<s16<u32', data))
    # logging.debug(bitstruct.unpack('<u1<u2<u1<u2<u1<p9<s16<u32', data))
    # logging.debug(bitstruct.unpack('<p9<u1<u1<u2<u1<u2<s16<u32', data))
    # logging.debug(pp.pformat(base_type(names['Relay']).padded_members()))
    logging.debug(pp.pformat(base_type(names['Relay']).unpack(data)))

    return

    indent_level = '    '
    for tag, items in objects.items():
        for die in items:
            logging.debug('{}: {}'.format(tag, die.attributes['DW_AT_name'].value.decode('utf-8')))

            if die.tag == 'DW_TAG_structure_type':
                logging.debug(indent_level + 'DIE tag=%s' % die.tag)
                try:
                    logging.debug(indent_level + '  Sibling: ' + str(
                        die.attributes['DW_AT_sibling'].value))
                except KeyError:
                    logging.debug(indent_level + '  KeyError on DW_AT_sibling')
                    pass
                logging.debug(indent_level + '  Offset: ' + str(die.offset))
                logging.debug(indent_level + '  File: ' + str(
                    die.attributes['DW_AT_decl_file'].value))
                logging.debug(indent_level + '  Line: ' + str(
                    die.attributes['DW_AT_decl_line'].value))
                logging.debug(indent_level + '  Attributes: ' + str(die.attributes))
                logging.debug(indent_level + '  DIE: ' + str(dir(die)))
                die_info_rec_structure_type(die, indent_level + '    ')
            # else:
            # logging.debug(indent_level + 'DIE tag=%s' % die.tag)
            elif die.tag == 'DW_TAG_typedef':
                logging.debug(indent_level + 'DIE tag=%s' % die.tag)
                try:
                    logging.debug(indent_level + '  Name: ' + str(
                        die.attributes['DW_AT_name'].value.decode('utf-8')))
                except KeyError:
                    logging.debug(indent_level + '  KeyError on DW_AT_name')
                    pass
                logging.debug(indent_level + '  Offset: ' + str(die.offset))
                file = die.attributes.get('DW_AT_decl_file', None)
                if file is not None:
                    file = file.value
                else:
                    file = 'not found'
                logging.debug(indent_level + '  File: ' + str(file))
                line = die.attributes.get('DW_AT_decl_line', None)
                if line is not None:
                    line = line.value
                else:
                    line = 'not found'
                logging.debug(indent_level + '  Line: ' + str(line))
                logging.debug(indent_level + '  Type: ' + str(die.attributes['DW_AT_type'].value))
                logging.debug(indent_level + '  describe_attr_value(Type): ' +
                      describe_attr_value(die.attributes['DW_AT_type'],
                                          die,
                                          die.cu.cu_offset))
                logging.debug(indent_level + '  Attributes: ' + str(die.attributes))
                logging.debug(indent_level + '  DIE: ' + str(dir(die)))
            elif die.tag == 'DW_TAG_variable':
                logging.debug(indent_level + 'DIE tag=%s' % die.tag)
            else:
                logging.debug(indent_level + 'DIE tag=%s' % die.tag)
                logging.debug(indent_level + '  Offset: ' + str(die.offset))
                for attribute, value in die.attributes.items():
                    logging.debug(indent_level + '  {}: {}'
                          .format(attribute,
                                  describe_attr_value(die.attributes[attribute],
                                                      die,
                                                      die.cu.cu_offset)))


def die_info_rec(die, indent_level='    ', objects=None):
    """ A recursive function for showing information about a DIE and its
        children.
    """
    logging.debug(indent_level + 'DIE tag=%s, name=%s' % (die.tag, die.attributes.get('DW_AT_name', None)))
    if objects is not None and die.tag in objects.keys():
        objects[die.tag].append(die)
    child_indent = indent_level + '  '
    for child in die.iter_children():
        die_info_rec(child, child_indent, objects)


if __name__ == '__main__':
    for filename in sys.argv[1:]:
        names, variables, bits_per_byte = process_file(filename)
        testit(names, variables)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'nvview.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(957, 598)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.horizontalLayout.setContentsMargins(-1, -1, 0, -1)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.module_to_nv_button = QtWidgets.QPushButton(Form)
        self.module_to_nv_button.setObjectName("module_to_nv_button")
        self.horizontalLayout.addWidget(self.module_to_nv_button)
        self.write_to_module_button = QtWidgets.QPushButton(Form)
        self.write_to_module_button.setObjectName("write_to_module_button")
        self.horizontalLayout.addWidget(self.write_to_module_button)
        self.write_to_file_button = QtWidgets.QPushButton(Form)
        self.write_to_file_button.setObjectName("write_to_file_button")
        self.horizontalLayout.addWidget(self.write_to_file_button)
        self.write_to_value_set_file_button = QtWidgets.QPushButton(Form)
        self.write_to_value_set_file_button.setObjectName("write_to_value_set_file_button")
        self.horizontalLayout.addWidget(self.write_to_value_set_file_button)
        self.write_to_auto_parameters_button = QtWidgets.QPushButton(Form)
        self.write_to_auto_parameters_button.setObjectName("write_to_auto_parameters_button")
        self.horizontalLayout.addWidget(self.write_to_auto_parameters_button)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.read_from_value_set_file_button = QtWidgets.QPushButton(Form)
        self.read_from_value_set_file_button.setObjectName("read_from_value_set_file_button")
        self.horizontalLayout.addWidget(self.read_from_value_set_file_button)
        self.read_from_file_button = QtWidgets.QPushButton(Form)
        self.read_from_file_button.setObjectName("read_from_file_button")
        self.horizontalLayout.addWidget(self.read_from_file_button)
        self.read_from_module_button = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.read_from_module_button.sizePolicy().hasHeightForWidth())
        self.read_from_module_button.setSizePolicy(sizePolicy)
        self.read_from_module_button.setObjectName("read_from_module_button")
        self.horizontalLayout.addWidget(self.read_from_module_button)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.label = QtWidgets.QLabel(Form)
        self.label.setObjectName("label")
        self.horizontalLayout_2.addWidget(self.label)
        self.access_level_password = QtWidgets.QLineEdit(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.access_level_password.sizePolicy().hasHeightForWidth())
        self.access_level_password.setSizePolicy(sizePolicy)
        self.access_level_password.setEchoMode(QtWidgets.QLineEdit.Password)
        self.access_level_password.setObjectName("access_level_password")
        self.horizontalLayout_2.addWidget(self.access_level_password)
        self.access_level = QtWidgets.QComboBox(Form)
        self.access_level.setObjectName("access_level")
        self.horizontalLayout_2.addWidget(self.access_level)
        self.set_access_level = QtWidgets.QPushButton(Form)
        self.set_access_level.setObjectName("set_access_level")
        self.horizontalLayout_2.addWidget(self.set_access_level)
        self.current_access_level = Enum(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.current_access_level.sizePolicy().hasHeightForWidth())
        self.current_access_level.setSizePolicy(sizePolicy)
        self.current_access_level.setProperty("label_visible", False)
        self.current_access_level.setObjectName("current_access_level")
        self.horizontalLayout_2.addWidget(self.current_access_level)
        self.line = QtWidgets.QFrame(Form)
        self.line.setFrameShape(QtWidgets.QFrame.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.horizontalLayout_2.addWidget(self.line)
        self.auto_read = QtWidgets.QCheckBox(Form)
        self.auto_read.setChecked(True)
        self.auto_read.setObjectName("auto_read")
        self.horizontalLayout_2.addWidget(self.auto_read)
        self.enforce_range_limits_check_box = QtWidgets.QCheckBox(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.enforce_range_limits_check_box.sizePolicy().hasHeightForWidth())
        self.enforce_range_limits_check_box.setSizePolicy(sizePolicy)
        self.enforce_range_limits_check_box.setChecked(True)
        self.enforce_range_limits_check_box.setObjectName("enforce_range_limits_check_box")
        self.horizontalLayout_2.addWidget(self.enforce_range_limits_check_box)
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.status_label = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.status_label.sizePolicy().hasHeightForWidth())
        self.status_label.setSizePolicy(sizePolicy)
        self.status_label.setText("")
        self.status_label.setObjectName("status_label")
        self.horizontalLayout_3.addWidget(self.status_label)
        self.label_2 = QtWidgets.QLabel(Form)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_3.addWidget(self.label_2)
        self.diff_reference_column = QtWidgets.QComboBox(Form)
        self.diff_reference_column.setObjectName("diff_reference_column")
        self.horizontalLayout_3.addWidget(self.diff_reference_column)
        self.verticalLayout_2.addLayout(self.horizontalLayout_3)
        self.searchbox = SearchBox(Form)
        self.searchbox.setObjectName("searchbox")
        self.verticalLayout_2.addWidget(self.searchbox)
        self.tree_view = TreeView(Form)
        self.tree_view.setAlternatingRowColors(True)
        self.tree_view.setObjectName("tree_view")
        self.verticalLayout_2.addWidget(self.tree_view)
        self.verticalLayout.addLayout(self.verticalLayout_2)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.module_to_nv_button.setText(_translate("Form", "Inverter To NV"))
        self.write_to_module_button.setText(_translate("Form", "To Inverter"))
        self.write_to_file_button.setText(_translate("Form", "To File"))
        self.write_to_value_set_file_button.setText(_translate("Form", "To Value Set File"))
        self.write_to_auto_parameters_button.setText(_translate("Form", "To Auto Parameters"))
        self.read_from_value_set_file_button.setText(_translate("Form", "From Value Set File"))
        self.read_from_file_button.setText(_translate("Form", "From File"))
        self.read_from_module_button.setText(_translate("Form", "From Inverter"))
        self.label.setText(_translate("Form", "Access Level:"))
        self.set_access_level.setText(_translate("Form", "Set"))
        self.auto_read.setText(_translate("Form", "Auto Read"))
        self.enforce_range_limits_check_box.setText(_translate("Form", "Enforce Limits Locally"))
        self.label_2.setText(_translate("Form", "Diff Reference Column:"))
from epyqlib.searchbox import SearchBox
from epyqlib.treeview import TreeView
from epyqlib.widgets.enum import Enum



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'check.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(653, 375)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(Form)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.value = QtWidgets.QCheckBox(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.value.sizePolicy().hasHeightForWidth())
        self.value.setSizePolicy(sizePolicy)
        self.value.setText("")
        self.value.setObjectName("value")
        self.horizontalLayout.addWidget(self.value)
        self.label = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.horizontalLayout_2.addLayout(self.horizontalLayout)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "TextLabel"))



from PyQt5 import QtWidgets

import epyqlib.faultlogview_ui
import epyqlib.utils.qt


# See file COPYING in this source tree
__copyright__ = 'Copyright 2018, EPC Power Corp.'
__license__ = 'GPLv2+'


class FaultLogView(QtWidgets.QWidget):
    def __init__(self, parent=None, in_designer=False):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        self.ui = epyqlib.faultlogview_ui.Ui_Form()
        self.ui.setupUi(self)

        view = self.ui.tree_view
        view.setSelectionBehavior(view.SelectRows)
        view.setSelectionMode(view.ExtendedSelection)

        self.model = None

        self.ui.clear_button.clicked.connect(self.clear)

    def set_model(self, model):
        self.model = model

        self.ui.tree_view.setModel(model)

    def clear(self):
        answer = QtWidgets.QMessageBox.question(
            self,
            'Clear Log',
            'Are you sure you want to clear the log?',
        )

        if answer != QtWidgets.QMessageBox.Yes:
            return

        children = list(self.model.root.children)
        for child in children:
            self.model.root.remove_child(child=child)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'variableselectionview.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(400, 300)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.searchbox = SearchBox(Form)
        self.searchbox.setObjectName("searchbox")
        self.gridLayout.addWidget(self.searchbox, 0, 0, 1, 1)
        self.tree_view = QtWidgets.QTreeView(Form)
        self.tree_view.setAlternatingRowColors(True)
        self.tree_view.setObjectName("tree_view")
        self.gridLayout.addWidget(self.tree_view, 1, 0, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
from epyqlib.searchbox import SearchBox



import struct

import attr
import siphash


struct_format = '<I16sHf'


def high_32(value):
    return (value >> 32) & 0xffffffff


def low_32(value):
    return value & 0xffffffff


@attr.s(frozen=True)
class Uint64:
    value = attr.ib()

    @classmethod
    def ensure(cls, value):
        if isinstance(value, cls):
            return value

        return cls(value=value)

    def low_32(self):
        return low_32(self.value)

    def high_32(self):
        return high_32(self.value)

    def to_bytes(self, byteorder='little'):
        return self.value.to_bytes(8, byteorder)


@attr.s(frozen=True)
class Tag64:
    value = attr.ib(converter=Uint64.ensure)


@attr.s(frozen=True)
class Key64:
    value = attr.ib(converter=Uint64.ensure)
    level = attr.ib()


# TODO: make it do something
# def generate_key():
#     pass


def pack(serial_number, parameter_uuid, meta_index, value):
    return struct.pack(
        struct_format,
        int(serial_number),  # TODO: make sure it doesn't lose info?
        parameter_uuid.bytes,
        meta_index,
        value,
    )


def build_tag(key, message):
    output_bytes = siphash.half_siphash_64(
        key=key.to_bytes(byteorder='little'),
        data=message,
    )

    tag = int.from_bytes(output_bytes, byteorder='little')

    return tag


def create_tag(key_value, serial_number, parameter_uuid, meta_index, value):
    message = pack(
        serial_number=serial_number,
        parameter_uuid=parameter_uuid,
        meta_index=meta_index,
        value=value,
    )

    tag_value = build_tag(key=key_value, message=message)

    return Tag64(value=tag_value)



import inspect
import itertools

import attr
import pytest
import PyQt5.QtCore
import PyQt5.QtGui

import epyqlib.tests.common
import epyqlib.utils.qt


@epyqlib.utils.qt.pyqtify()
@attr.s
class P(PyQt5.QtCore.QObject):
    a = attr.ib()
    b = attr.ib()
    c = attr.ib()

    def __attrs_post_init__(self):
        super().__init__()
        self.get = None
        self.set = None

    @PyQt5.QtCore.pyqtProperty('PyQt_PyObject')
    def pyqtify_c(self):
        x = epyqlib.utils.qt.pyqtify_get(self, 'c')
        self.get = True
        return x

    @pyqtify_c.setter
    def pyqtify_c(self, value):
        self.set = True
        return epyqlib.utils.qt.pyqtify_set(self, 'c', value)


def test_types(qtbot):
    p = P(a=1, b=2, c=3)

    # TODO: this would be nice but all ideas so far are ugly
    # assert isinstance(P.a, attr.Attribute)

    assert isinstance(attr.fields(P).a, attr.Attribute)
    assert isinstance(inspect.getattr_static(p, 'a'), property)
    assert isinstance(inspect.getattr_static(p, 'c'), PyQt5.QtCore.pyqtProperty)


def assert_attrs_as_expected(x, values):
    assert attr.asdict(x) == {
        k: tuple(itertools.chain((v.initial,), v.expected))[-1]
        for k, v in values.items()
    }


def test_overall(qtbot):
    values = {
        'a': epyqlib.tests.common.Values(
            initial=1,
            input=[12, 12, 13],
            expected=[12, 13],
        ),
        'b': epyqlib.tests.common.Values(
            initial=2,
            input=[42, 42, 37],
            expected=[42, 37],
        ),
        'c': epyqlib.tests.common.Values(
            initial=3,
            input=[4],
            expected=[4],
        ),
    }

    p = P(**{k: v.initial for k, v in values.items()})
    fields = attr.fields(P)
    assert len(fields) == len(values)

    signals = epyqlib.utils.qt.pyqtify_signals(p)
    for name, v in values.items():
        getattr(signals, name).connect(v.collect)

    for name, v in values.items():
        for value in v.input:
            setattr(p, name, value)

    for name, v in values.items():
        assert tuple(v.expected) == tuple(v.collected)

    assert_attrs_as_expected(p, values)

    p.c = object()
    assert p.pyqtify_c is p.c

    p.get = False
    p.c
    assert p.get

    p.set = False
    p.c = 0
    assert p.set


def test_independence(qtbot):
    ad = {'a': 1, 'b': 2, 'c': 3}
    a = P(**ad)

    bd = {'a': 10, 'b': 20, 'c': 30}
    b = P(**bd)

    assert(attr.asdict(a) == ad)
    assert(attr.asdict(b) == bd)


@epyqlib.utils.qt.pyqtify()
@attr.s
class Q(PyQt5.QtCore.QObject):
    a = attr.ib()
    b = attr.ib()

    def __attrs_post_init__(self):
        super().__init__()
        self.get = None
        self.set = None

    @PyQt5.QtCore.pyqtProperty('PyQt_PyObject')
    def pyqtify_b(self):
        return epyqlib.utils.qt.pyqtify_get(self, 'b')

    @pyqtify_b.setter
    def pyqtify_b(self, value):
        # self.a = min(self.a, value)
        if value < self.a:
            self.a = value

        epyqlib.utils.qt.pyqtify_set(self, 'b', value)


def test_property_cross_effect(qtbot):
    values = {
        'a': epyqlib.tests.common.Values(
            initial=10,
            input=[],
            expected=[9],
        ),
        'b': epyqlib.tests.common.Values(
            initial=20,
            input=[30, 10, 9, 20],
            expected=[30, 10, 9, 20],
        ),
    }

    p = Q(**{k: v.initial for k, v in values.items()})
    fields = attr.fields(Q)
    assert len(fields) == len(values)

    signals = epyqlib.utils.qt.pyqtify_signals(p)
    for name, v in values.items():
        getattr(signals, name).connect(v.collect)

    for name, v in values.items():
        for value in v.input:
            setattr(p, name, value)

    for name, v in values.items():
        assert tuple(v.expected) == tuple(v.collected)

    assert_attrs_as_expected(p, values)


def test_pyqtified_name():
    assert Q.__name__ == 'Q'


def test_pyqtified_module():
    class C:
        pass

    assert Q.__module__ == C.__module__


def test_(qtbot):
    q = Q(a=1, b=2)

    signals = epyqlib.utils.qt.pyqtify_signals(q)

    # TODO: Actually assert they are 'the same'.  Until we know how to
    #       just access them to make sure they are available
    signals._pyqtify_signal_a
    signals.a
    signals['a']


def test_resolve_index_to_model():
    model = PyQt5.QtCore.QStringListModel()

    back_proxy = PyQt5.QtCore.QSortFilterProxyModel()
    back_proxy.setSourceModel(model)

    middle_proxy = PyQt5.QtCore.QSortFilterProxyModel()
    middle_proxy.setSourceModel(back_proxy)

    proxy = PyQt5.QtCore.QSortFilterProxyModel()
    proxy.setSourceModel(middle_proxy)

    view = PyQt5.QtWidgets.QListView()
    view.setModel(proxy)

    assert (
        epyqlib.utils.qt.resolve_models(model=view.model())
        == [proxy, middle_proxy, back_proxy, model]
    )

    strings = ['a', 'b', 'c']
    model.setStringList(strings)
    assert model.rowCount() == 3
    assert model.stringList() == strings

    proxy_first_index = proxy.index(0, 0, PyQt5.QtCore.QModelIndex())

    model_first_index = model.index(0, 0, PyQt5.QtCore.QModelIndex())

    target_data = model.data(model_first_index, PyQt5.QtCore.Qt.DisplayRole)

    assert target_data == 'a'

    with pytest.raises(epyqlib.utils.qt.TargetModelNotReached):
        epyqlib.utils.qt.resolve_index_to_model(
            index=proxy_first_index,
            target=object(),
        )

    index = epyqlib.utils.qt.resolve_index_to_model(
        index=proxy_first_index,
    )

    found_data = model.data(index, PyQt5.QtCore.Qt.DisplayRole)

    assert found_data == target_data
    assert index.model() == model

    proxy_index = epyqlib.utils.qt.resolve_index_from_model(
        model=model,
        view=view,
        index=model_first_index,
    )

    assert (
        proxy.data(proxy_index, PyQt5.QtCore.Qt.DisplayRole)
        == model.data(model_first_index, PyQt5.QtCore.Qt.DisplayRole)
    )


def test_attrs_no_recurse_in_init():
    @epyqlib.utils.qt.pyqtify()
    @attr.s
    class Child:
        a = attr.ib(default=42)

    @epyqlib.utils.qt.pyqtify()
    @attr.s
    class Parent:
        child = attr.ib(default=attr.Factory(Child))

    p = Parent()

    assert p.child == Child()


def test_signal_independence():
    class C:
        a = epyqlib.utils.qt.Signal(int)
        b = epyqlib.utils.qt.Signal(int)

    value_checkers = {
        'a': epyqlib.tests.common.Values(
            initial=None,
            input=[1, 2, 3],
            expected=[1, 2, 3],
        ),
        'b': epyqlib.tests.common.Values(
            initial=None,
            input=[10, 20, 30],
            expected=[10, 20, 30],
        ),
    }

    c = C()
    for name, checker in value_checkers.items():
        getattr(c, name).connect(checker.collect)

    for name, checker in value_checkers.items():
        for value in checker.input:
            getattr(c, name).emit(value)

    for name, checker in value_checkers.items():
        assert tuple(checker.expected) == tuple(checker.collected)


def test_signal_chaining():
    class C:
        a = epyqlib.utils.qt.Signal(int)
        b = epyqlib.utils.qt.Signal(int)

    input = [1, 2, 3]

    value_checkers = {
        'a': epyqlib.tests.common.Values(
            initial=None,
            input=input,
            expected=input,
        ),
        'b': epyqlib.tests.common.Values(
            initial=None,
            input=None,
            expected=input,
        ),
    }

    c = C()
    c.a.connect(c.b)

    for name, checker in value_checkers.items():
        getattr(c, name).connect(checker.collect)

    for value in value_checkers['a'].input:
        c.a.emit(value)

    for name, checker in value_checkers.items():
        assert tuple(checker.expected) == tuple(checker.collected)


def test_signal_to_pyqtslot():
    class Signal:
        signal = epyqlib.utils.qt.Signal()

    class Slot(PyQt5.QtCore.QObject):
        @PyQt5.QtCore.pyqtSlot()
        def slot(self):
            pass

    s = Signal()
    q = Slot()

    s.signal.connect(q.slot)


# def test_signal_repr():
#     class C:
#         signal = epyqlib.utils.qt.Signal()
#
#     c = C()
#
#     expected = (
#         '<bound PYQT_SIGNAL signal of _SignalQObject object at 0x'
#         ' of C object at 0x'
#     )
#
#     actual = repr(c.signal)
#     print(actual)
#     actual = actual.split()
#
#     for index in (-1, -6):
#         actual[index] = actual[index][:2]
#
#     actual = ' '.join(actual)
#
#     assert actual == expected


@attr.s(slots=True)
class DiffProxy:
    model = attr.ib()
    proxy = attr.ib()

    def visit_all(self, visitor):
        for row in range(self.proxy.rowCount()):
            for column in range(self.proxy.columnCount()):
                visitor(row, column)

    def lists(self, fill=None):
        return [
            [fill for _ in range(self.proxy.rowCount())]
            for _ in range(self.proxy.columnCount())
        ]

    def role_lists(self, fill=None):
        return {
            role: self.lists(fill=fill)
            for role in self.roles()
        }

    def roles(self):
        return {*self.proxy.diff_highlights, *self.proxy.reference_highlights}

    def collect(self):
        results = {
            self.proxy.diff_role: self.lists(),
            **{
                role: self.lists()
                for role in self.roles()
            },
        }

        def collect(row, column, collected=results):
            for role, lists in collected.items():
                index = self.proxy.index(
                    row,
                    column,
                    PyQt5.QtCore.QModelIndex(),
                )

                lists[row][column] = self.proxy.data(index, role)

        self.visit_all(visitor=collect)

        return results


@pytest.fixture
def diff_proxy_test_model():
    rows = 4
    columns = 4

    model = PyQt5.QtGui.QStandardItemModel(rows, columns)

    for row in range(rows):
        for column in range(columns):
            model.setItem(row, column, PyQt5.QtGui.QStandardItem())

    proxy = epyqlib.utils.qt.DiffProxyModel(
        columns=range(1, rows),
        diff_highlights={
            PyQt5.QtCore.Qt.ItemDataRole.BackgroundRole: (
                PyQt5.QtGui.QColor('orange')
            )
        },
        reference_highlights={
            PyQt5.QtCore.Qt.ItemDataRole.BackgroundRole: (
                PyQt5.QtGui.QColor('green')
            )
        },
    )
    proxy.setSourceModel(model)

    return DiffProxy(model=model, proxy=proxy)


def test_diffproxymodel_no_reference_column(diff_proxy_test_model):
    diff_proxy_test_model.proxy.reference_column = None

    results = diff_proxy_test_model.collect()
    results = {
        role: results[role]
        for role in diff_proxy_test_model.proxy.diff_highlights
    }

    expected = diff_proxy_test_model.role_lists()

    assert expected == results


def test_diffproxymodel_some_differences(diff_proxy_test_model):
    reference_column = 2
    root = diff_proxy_test_model.model.invisibleRootItem()
    rows = root.rowCount()
    columns = root.columnCount()

    for row in range(rows):
        for column in range(columns):
            root.child(row, column).setData(row)

    diff = {
        (2, 1): 42,
        (3, 3): 42,
    }

    expected = diff_proxy_test_model.role_lists()

    for (row, column), value in diff.items():
        item = root.child(row, column)
        item.setData(
            value,
            diff_proxy_test_model.proxy.diff_role,
        )

        diff_highlights = diff_proxy_test_model.proxy.diff_highlights.items()
        for role, highlight in diff_highlights:
            expected[role][row][column] = highlight

    reference_highlights = (
        diff_proxy_test_model.proxy.reference_highlights.items()
    )
    for row in range(rows):
        for role, highlight in reference_highlights:
            expected[role][row][reference_column] = highlight

    diff_proxy_test_model.proxy.reference_column = reference_column

    results = diff_proxy_test_model.collect()
    results = {
        role: results[role]
        for role in diff_proxy_test_model.proxy.diff_highlights
    }

    assert expected == results


@attr.s(slots=True)
class ChangedData:
    start = attr.ib()
    end = attr.ib()
    roles = attr.ib()


@attr.s(slots=True)
class DataChanges:
    # waiting on https://github.com/python-attrs/attrs/pull/420
    __weakref__ = attr.ib(
        repr=False,
        cmp=False,
        hash=False,
        init=False,
    )
    changes = attr.ib(factory=list)

    def collect(self, start, end, roles):
        self.changes.append(ChangedData(
            start=start,
            end=end,
            roles=roles,
        ))

    def results(self, container):
        for changed_data in self.changes:
            start = changed_data.start
            end = changed_data.end

            for row in range(start.row(), end.row() + 1):
                for column in range(start.column(), end.column() + 1):
                    for role in changed_data.roles:
                        if role in container:
                            container[role][row][column] = True

        return container


def test_diffproxymodel_all_changed(diff_proxy_test_model):
    changes = DataChanges()

    diff_proxy_test_model.proxy.dataChanged.connect(changes.collect)
    diff_proxy_test_model.proxy.all_changed()

    expected = diff_proxy_test_model.role_lists(fill=False)

    root = diff_proxy_test_model.model.invisibleRootItem()

    for role in diff_proxy_test_model.proxy.diff_highlights:
        for row in range(root.rowCount()):
            for column in diff_proxy_test_model.proxy.columns:
                expected[role][row][column] = True

    assert expected == changes.results(
        container=diff_proxy_test_model.role_lists(fill=False),
    )


def test_diffproxymodel_edit_reference_emits(diff_proxy_test_model):
    diff_proxy_test_model.proxy.reference_column = 2
    changes = DataChanges()

    proxy = diff_proxy_test_model.proxy
    proxy.dataChanged.connect(changes.collect)
    row = 0
    assert proxy.setData(
        index=proxy.index(row, proxy.reference_column),
        value=42,
        role=proxy.diff_role,
    )

    expected = diff_proxy_test_model.role_lists(fill=False)
    for role in proxy.diff_highlights:
        for column in proxy.columns:
            expected[role][row][column] = True

    assert expected == changes.results(
        container=diff_proxy_test_model.role_lists(fill=False),
    )


def test_diffproxymodel_edit_nonreference_emits(diff_proxy_test_model):
    diff_proxy_test_model.proxy.reference_column = 2
    edit_column = 1
    changes = DataChanges()

    proxy = diff_proxy_test_model.proxy
    proxy.dataChanged.connect(changes.collect)
    row = 0
    assert proxy.setData(
        index=proxy.index(row, edit_column),
        value=42,
        role=proxy.diff_role,
    )

    expected = diff_proxy_test_model.role_lists(fill=False)
    for role in proxy.diff_highlights:
        expected[role][row][edit_column] = True

    assert expected == changes.results(
        container=diff_proxy_test_model.role_lists(fill=False),
    )



# This file has been generated


build_system = 'AppVeyor'
build_id = '28463587'
build_number = '385'
build_version = '1.0.385'
job_id = 'lqve17y8c2nlvem5'
job_url = 'https://ci.appveyor.com/project/KyleAltendorf/stlib/build/job/lqve17y8c2nlvem5'



import pathlib

import attr
import pytest

import epyqlib.tests.common
import epyqlib.hildevice


@pytest.fixture
def factory_definition():
    return epyqlib.hildevice.Definition.loadp(
        epyqlib.tests.common.devices['factory'],
    )


def test_definition_format_version_validator(factory_definition):
    with pytest.raises(epyqlib.hildevice.FormatVersionError):
        attr.evolve(factory_definition, format_version=[2])


def test_definition_load(factory_definition):
    assert factory_definition.base_path.is_dir()
    assert factory_definition.can_path.exists()
    assert factory_definition.access_level_path is not None
    assert factory_definition.access_password_path is not None


def test_definition_loads():
    path = pathlib.Path(epyqlib.tests.common.devices['factory'])
    with open(path) as f:
        s = f.read()

    epyqlib.hildevice.Definition.loads(
        s=s,
        base_path=path.parents[0],
    )


def test_load():
    device = epyqlib.hildevice.Device(
        definition_path=epyqlib.tests.common.devices['factory'],
    )
    device.load()


def test_device_reload():
    device = epyqlib.hildevice.Device(
        definition_path=epyqlib.tests.common.devices['factory'],
    )
    device.load()

    with pytest.raises(epyqlib.hildevice.AlreadyLoadedError):
        device.load()



import collections
import itertools
import logging
import os
import signal
import time

import attr
import canmatrix.formats
import click
import docx
import docx.enum.section
import docx.enum.table
import docx.enum.text
import docx.oxml
import docx.oxml.ns
import docx.shared
import docx.table
import lxml.etree

import epyqlib.utils.general

# See file COPYING in this source tree
__copyright__ = 'Copyright 2017, EPC Power Corp.'
__license__ = 'GPLv2+'


logger = logging.getLogger(__name__)


# https://github.com/python-openxml/python-docx/issues/322#issuecomment-265018856
def set_repeat_table_header(row):
    tr = row._tr
    trPr = tr.get_or_add_trPr()
    tblHeader = docx.oxml.OxmlElement('w:tblHeader')
    tblHeader.set(docx.oxml.ns.qn('w:val'), "true")
    trPr.append(tblHeader)
    return row


# https://github.com/python-openxml/python-docx/issues/245#issuecomment-208476933
def prevent_row_breaks(table):
    tags = table._element.xpath('//w:tr')
    rows = len(tags)
    for row in range(0,rows):
        tag = tags[row]                     # Specify which <w:r> tag you want
        child = docx.oxml.OxmlElement('w:cantSplit')  # Create arbitrary tag
        tag.append(child)                   # Append in the new tag


def w(s):
    return '{http://schemas.openxmlformats.org/wordprocessingml/2006/main}' + s


def shd_element(fill):
    e = lxml.etree.Element(w('shd'))
    # <w:shd w:val="clear" w:color="auto" w:fill="D9D9D9" w:themeFill="background1" w:themeFillShade="D9"/>

    e.attrib[w('fill')] = fill

    return e


def shade(cell, fill):
    cell._tc.tcPr.append(shd_element(fill=fill))


@attr.s
class Table:
    title = attr.ib()
    headings = attr.ib()
    widths = attr.ib()
    total_width = attr.ib(default=10)
    comment = attr.ib(default='')
    rows = attr.ib(default=attr.Factory(list))

    def fill_docx(
            self,
            table,
            title_style=None,
            heading_style=None,
            contents_style=None,
    ):
        table.alignment = docx.enum.table.WD_TABLE_ALIGNMENT.CENTER
        # table.autofit = True
        # table.style = 'CAN Table Base'

        title_present = self.title is not None and len(self.title) > 0

        if title_present:
            row = table.add_row()
            row.cells[0].text = self.title
            title_paragraph = row.cells[0].paragraphs[0]
            title_paragraph.paragraph_format.keep_with_next = True
            if title_style is not None:
                title_paragraph.style = title_style
            shade(row.cells[0], fill="000000")

        if len(self.comment) > 0:
            row = table.add_row()
            row.cells[0].merge(row.cells[-1])
            row.cells[0].text = self.comment
            row.cells[0].paragraphs[0].paragraph_format.keep_with_next = True

        row = table.add_row()
        for cell, heading in zip(row.cells, self.headings):
            cell.text = heading
            if heading_style is not None:
                cell.paragraphs[0].style = heading_style
            cell.paragraphs[0].paragraph_format.keep_with_next = True

            if not title_present:
                shade(cell, fill="000000")

        shadings = (
            {'fill': 'D9D9D9'},
            None,
        )

        shadings_iterator = itertools.cycle(shadings)
        if not title_present:
            next(shadings_iterator)

        zipped = zip(self.rows, shadings_iterator)
        total = len(self.rows)
        checkpoint = time.monotonic()
        n = 10
        for i, (r, shading) in enumerate(zipped):
            if i > 0 and i % n == 0:
                now = time.monotonic()
                delta = now - checkpoint
                checkpoint = now
                print('{} / {} ({:.3})'.format(i, total, delta / n))

            row = table.add_row()
            for cell, text in zip(row.cells, r):
                cell.text = str(text)
                if shading is not None:
                    shade(cell, **shading)
                if contents_style is not None:
                    cell.paragraphs[0].style = contents_style
                cell.paragraphs[0].paragraph_format.keep_with_next = True

        remaining_width = (
            self.total_width - sum(w for w in self.widths if w is not None)
        )
        each_width = remaining_width / sum(1 for w in self.widths if w is None)
        widths = [each_width if w is None else w for w in self.widths]
        widths = [
            w if w is None else docx.shared.Inches(w)
            for w in widths
        ]
        for column, width in zip(table.columns, widths):
            column.width = width

        total = len(self.rows)
        checkpoint = time.monotonic()
        n = 10
        for i, row in enumerate(table.rows):
            if i > 0 and i % n == 0:
                now = time.monotonic()
                delta = now - checkpoint
                checkpoint = now
                print('{} / {} ({:.3})'.format(i, total, delta / n))

            for cell, width in zip(row.cells, widths):
                cell.width = width

        if title_present:
            table.rows[0].cells[0].merge(table.rows[0].cells[-1])


def id_string(id):
    return '0x{:08X}'.format(id)


def tabulate_signals(signals):
    rows = []

    for signal in signals:
        # if len(signal.unit) == 0 and signal.factor != 1:
            # table.append(('***',) * 5)
        startbit = signal.getStartbit()
        rows.append((
            '', '',
            signal.name,
            '{}/{}'.format(startbit % 8, startbit // 8),
            signal.size,
            signal.factor,
            signal.unit,
            '{}: {}'.format(signal.enumeration, signal.values) if
            signal.enumeration is not None else '',
        ))
        if len(signal.comment) > 0:
            rows.append((*(('',) * 7), signal.comment))

    return rows


def doc_signals(signals):
    rows = []

    for signal in signals:
        startbit = signal.getStartbit()
        enumeration = (
            signal.enumeration
            if signal.enumeration is not None
            else ''
        )
        rows.append((
            signal.name,
            '{}/{}'.format(startbit % 8, startbit // 8),
            signal.size,
            signal.factor,
            signal.unit,
            enumeration,
            signal.comment,
        ))

    return rows

@click.command()
@click.option('--can', '-c', type=click.File('rb'), required=True)
@click.option('--template', '-t', type=click.File('rb'), required=True)
@click.option('--output', '-o', type=click.File('wb'), required=True)
@click.option('--verbose', '-v', count=True)
def main(can, template, output, verbose):
    if verbose >= 1:
        logger.setLevel(logging.DEBUG)

    matrix, = canmatrix.formats.load(
        can,
        os.path.splitext(can.name)[1].lstrip('.')
    ).values()

    mux_table_header = (
        'Mux Name', 'Mux ID', 'Name', 'Start', 'Length', 'Factor', 'Units',
        'Enumeration', 'Comment'
    )
    mux_table = epyqlib.utils.general.TextTable()
    mux_table.append(mux_table_header)
    mux_table_header = mux_table_header[2:]

    frame_table_header = (
        'Frame', 'ID', 'Name', 'Start', 'Length', 'Factor', 'Units',
        'Enumeration', 'Comment'
    )

    frame_table = epyqlib.utils.general.TextTable()
    frame_table.append(frame_table_header)
    frame_table_header = frame_table_header[2:]

    widths = [0.625] * len(frame_table_header)
    widths[0] = 2
    widths[-2] = 1.5
    widths[-1] = None

    enum_table_header = ('Value', 'Name')

    frame_tables = []
    multiplex_tables = []
    enumeration_tables = []

    for frame in sorted(matrix.frames, key=lambda f: f.name):
        frame_table.append(
            frame.name,
            id_string(frame.id),
        )

        a_ft = Table(
            title='{} ({})'.format(frame.name, id_string(frame.id)),
            comment=frame.comment,
            headings=frame_table_header,
            widths=widths,
        )
        frame_tables.append(a_ft)

        mux_table.append(
            '{} ({})'.format(frame.name, id_string(frame.id)),
        )

        multiplex_signal = frame.signals[0]
        if multiplex_signal.multiplex is None:
            multiplex_signal = None

        if multiplex_signal is None:
            signals = sorted(frame.signals, key=lambda s: s.name)
            table = tabulate_signals(signals)
            frame_table.extend(table)

            a_ft.rows.extend(doc_signals(signals))
        else:
            signals = (multiplex_signal,)
            table = tabulate_signals(signals)
            frame_table.extend(table)
            a_ft.rows.extend(doc_signals(signals))
            # multiplex_signal.values = {
            #     int(k): v for k, v in multiplex_signal.values.items()
            # }

            for value, name in sorted(multiplex_signal.values.items()):
                if value == 0:
                    continue

                a_mt = Table(
                    title='{} ({}) - {} ({})'.format(
                        frame.name,
                        id_string(frame.id),
                        name,
                        value
                    ),
                    # comment=frame.comment,
                    headings=mux_table_header,
                    widths=widths,
                )
                multiplex_tables.append(a_mt)

                mux_table.append(
                    '',
                    value,
                    name,
                )

                signals = tuple(
                    s for s in frame.signals
                    if s.multiplex == value and s.name != 'ReadParam_command'
                )

                mux_table.extend(tabulate_signals(sorted(
                    signals, key=lambda s: s.name)))

                a_mt.rows.extend(doc_signals(signals))

    print('\n\n - - - - - - - Multiplexes\n')
    print(mux_table)

    print('\n\n - - - - - - - Frames\n')
    print(frame_table)

    enumeration_table = epyqlib.utils.general.TextTable()
    enumeration_table.append('Name', '', 'Value')
    widths = (1.5, None)
    for name, values in sorted(matrix.valueTables.items()):
        a_et = Table(
            title=name,
            headings=enum_table_header,
            widths=widths,
            total_width=5,
        )
        enumeration_tables.append(a_et)
        a_et.rows.extend(sorted(values.items()))

        enumeration_table.append(name)
        for i, s in sorted(values.items()):
            enumeration_table.append('', s, i)

    print('\n\n - - - - - - - Enumerations\n')
    print(enumeration_table)

    doc = docx.Document(template)

    table_sets = {
        'frames': frame_tables,
        'multiplexers': multiplex_tables,
        'enumerations': enumeration_tables,
    }
    for tag, tables in table_sets.items():
        full_tag = '<gen_{}>'.format(tag)
        for paragraph in doc.paragraphs:
            if paragraph.text.strip() == full_tag:
                break
            elif full_tag in paragraph.text:
                f = 'Tag {} found, expected as only text in paragraph: {}'
                raise Exception(f.format(full_tag, repr(paragraph.text)))
        else:
            raise Exception('Tag not found: {}'.format(full_tag))

        for table in tables:
            doc_table = doc.add_table(rows=0, cols=len(table.headings))
            doc_paragraph = doc.add_paragraph()

            paragraph._p.addprevious(doc_table._tbl)
            paragraph._p.addprevious(doc_paragraph._p)

            table.fill_docx(
                doc_table,
                title_style='CAN Table Title',
                heading_style='CAN Table Heading',
                contents_style='CAN Table Contents',
            )

        # TODO: Would rather delete the tag paragraph but that breaks the
        #       template's landscape page format for some reason
        # paragraph._p.getparent().remove(paragraph._p)
        paragraph.clear()
    doc.save(output)
    pass


def _entry_point():
    logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s')

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    return main()



import os
import textwrap

import click


@click.command()
@click.argument('target', type=click.File('w'), default='-')
def write_build_file(target):
    template = textwrap.dedent('''\
    # This file has been generated


    ''')

    target.write(template)

    values = {name: None for name in (
        'build_system',
        'build_id',
        'build_number',
        'build_version',
        'job_id',
        'job_url',
    )}

    if os.environ.get('APPVEYOR') == 'True':
        values['build_system'] = 'AppVeyor'
        mapping = {
            'build_id': 'APPVEYOR_BUILD_ID',
            'build_number': 'APPVEYOR_BUILD_NUMBER',
            'build_version': 'APPVEYOR_BUILD_VERSION',
            'job_id': 'APPVEYOR_JOB_ID',
        }
        values.update({k: os.environ[v] for k, v in mapping.items()})
        values['job_url'] = (
            'https://ci.appveyor.com/'
            'project/{account}/{slug}/build/job/{id}'.format(
                account=os.environ['APPVEYOR_ACCOUNT_NAME'],
                slug=os.environ['APPVEYOR_PROJECT_SLUG'],
                id=os.environ['APPVEYOR_JOB_ID'],
            )
        )

    for k, v in values.items():
        target.write('{} = {}\n'.format(k, repr(v)))



from PyQt5 import QtCore
from PyQt5 import QtWidgets

import epyqlib.searchbox_ui
import epyqlib.utils.qt


# See file COPYING in this source tree
__copyright__ = 'Copyright 2017, EPC Power Corp.'
__license__ = 'GPLv2+'


class SearchBox(QtWidgets.QWidget):
    filter_text_changed = QtCore.pyqtSignal(str)
    filter_requested = QtCore.pyqtSignal()

    search_text_changed = QtCore.pyqtSignal(str)
    search_requested = QtCore.pyqtSignal()

    def __init__(self, parent=None, in_designer=False):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        self._hide_search = False
        self._hide_filter = False

        self.ui = epyqlib.searchbox_ui.Ui_Form()
        self.ui.setupUi(self)

        self.hide_filter = True

        self.search_shortcut = None
        self.view = None

        self.ui.search_text.setPlaceholderText('Search...')
        self.ui.search_text.textChanged.connect(self.search_text_changed)
        self.ui.search_text.returnPressed.connect(self.search_requested)

        self.ui.filter_text.setPlaceholderText('Filter...')
        self.ui.filter_text.textChanged.connect(self.filter_text_changed)
        self.ui.filter_text.textChanged.connect(self.filter_requested)

    @QtCore.pyqtProperty(bool)
    def hide_search(self):
        return self._hide_search

    @hide_search.setter
    def hide_search(self, hide):
        self._hide_search = hide
        self.ui.search_text.setHidden(self.hide_search)

    @QtCore.pyqtProperty(bool)
    def hide_filter(self):
        return self._hide_filter

    @hide_filter.setter
    def hide_filter(self, hide):
        self._hide_filter = hide
        self.ui.filter_text.setHidden(self.hide_filter)

    def focus_search(self):
        self.ui.search_text.setFocus()

    @property
    def search_text(self):
        return self.ui.search_text.text()

    @search_text.setter
    def search_text(self, text):
        self.ui.search_text.setText(text)

    @property
    def filter_text(self):
        return self.ui.filter_text.text()

    @filter_text.setter
    def filter_text(self, text):
        self.ui.filter_text.setText(text)

    def connect_to_view(self, view, column):
        def _search():
            epyqlib.utils.qt.search_view(
                view=view,
                text=self.search_text,
                column=column,
            )
        self.search_requested.connect(_search)

        def _filter():
            view.model().setFilterWildcard(self.filter_text)
        self.filter_requested.connect(_filter)

        self.search_shortcut = QtWidgets.QShortcut(view)
        self.search_shortcut.setKey(QtCore.Qt.CTRL + QtCore.Qt.Key_F)
        self.search_shortcut.activated.connect(self.focus_search)
        self.search_shortcut.setContext(QtCore.Qt.WidgetWithChildrenShortcut)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'iopoint.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(795, 224)
        self.horizontalLayout = QtWidgets.QHBoxLayout(Form)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.status = Led(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.status.sizePolicy().hasHeightForWidth())
        self.status.setSizePolicy(sizePolicy)
        self.status.setProperty("label_visible", True)
        self.status.setObjectName("status")
        self.horizontalLayout.addWidget(self.status)
        self.set_label = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.set_label.sizePolicy().hasHeightForWidth())
        self.set_label.setSizePolicy(sizePolicy)
        self.set_label.setObjectName("set_label")
        self.horizontalLayout.addWidget(self.set_label)
        self.set = Check(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.set.sizePolicy().hasHeightForWidth())
        self.set.setSizePolicy(sizePolicy)
        self.set.setProperty("label_visible", False)
        self.set.setObjectName("set")
        self.horizontalLayout.addWidget(self.set)
        self.override = Check(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.override.sizePolicy().hasHeightForWidth())
        self.override.setSizePolicy(sizePolicy)
        self.override.setProperty("label_visible", False)
        self.override.setObjectName("override")
        self.horizontalLayout.addWidget(self.override)
        self.override_label = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.override_label.sizePolicy().hasHeightForWidth())
        self.override_label.setSizePolicy(sizePolicy)
        self.override_label.setObjectName("override_label")
        self.horizontalLayout.addWidget(self.override_label)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.set_label.setText(_translate("Form", "Set"))
        self.override_label.setText(_translate("Form", "Override"))
from epyqlib.widgets.check import Check
from epyqlib.widgets.led import Led



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'searchbox.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(765, 300)
        self.horizontalLayout = QtWidgets.QHBoxLayout(Form)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.search_text = QtWidgets.QLineEdit(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.search_text.sizePolicy().hasHeightForWidth())
        self.search_text.setSizePolicy(sizePolicy)
        self.search_text.setClearButtonEnabled(True)
        self.search_text.setObjectName("search_text")
        self.horizontalLayout.addWidget(self.search_text)
        spacerItem = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.filter_text = QtWidgets.QLineEdit(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.filter_text.sizePolicy().hasHeightForWidth())
        self.filter_text.setSizePolicy(sizePolicy)
        self.filter_text.setClearButtonEnabled(True)
        self.filter_text.setObjectName("filter_text")
        self.horizontalLayout.addWidget(self.filter_text)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
